//
//  Webserver.cs
//
//  Copyright (c) Lukas Lipka, 2007.
//

using System;
using System.IO;
using System.Net;
using System.Threading;
using System.Collections;
using System.Collections.Generic;

namespace Banshee.Webcast {

	public class Webserver {
		
		private Thread thread = null;
		private bool running = false;

		private HttpListener listener = null;
		private Hashtable mappings = null;
		private List<IBroker> brokers = new List<IBroker> ();

		public Webserver (string address, int port)
		{
			this.mappings = new Hashtable ();

			if (!HttpListener.IsSupported) {
				Log.Error (this, "HttpListener not supported! Bailing out...");
				Environment.Exit (1);
				return;
			}						

			listener = new HttpListener();
			listener.AuthenticationSchemes = AuthenticationSchemes.Negotiate;

			// FIXME: Access can be configured here. Currently we don't respect
			// the specified address, but use '+' as a wildcard to listen for all
			// connections on the specified port, regardless of the IP address passed
			// in.
			listener.Prefixes.Add (String.Format ("http://+:{1}/", address, port));

			Log.Debug (this, "Server started on {0}:{1}", address, port);

			// FIXME: We will most likely want to map this from outside of
			// the Webserver class itself.
			Map ("/", Directory.GetCurrentDirectory ());
			Map ("/test/foo", "/home/test/");
		}

		public void Register (IBroker broker)
		{
			Log.Debug (this, "Registered broker - {0}", broker.Name);
			brokers.Add (broker);
		}

		public void Map (string virtual_path, string local_path)
		{
			Log.Debug (this, "Mapped '{0}' to {1}", virtual_path, local_path);
			mappings [virtual_path] = local_path;
		}

		public string ResolveMap (string vpath)
		{
			string local_path = null;
			
			while (vpath.Length > 1) {
				local_path = (string)mappings [vpath];

				if (!String.IsNullOrEmpty (local_path)) {
					Log.Debug (this, "Resolved map '{0}' to '{1}'", vpath, local_path);
					return local_path;
				}
				
				vpath = vpath.Substring (0, vpath.LastIndexOf ('/'));
				
				Log.Debug (this, "Failed to resolve, trying with {0}", vpath);
			}

			return (string)mappings ["/"];
		}

		public void Start ()
		{
			listener.Start ();

			running = true;

			thread = new Thread (new ThreadStart (DoWork));
			thread.Start ();

			Log.Debug (this, "Started...");
		}

		public void Stop ()
		{
			if (!running)
				return;

			running = false;
			listener.Stop ();
			thread.Join (1000);

			Log.Debug (this, "Stopped...");
		}

		private void DoWork ()
		{
			while (running) {
				HttpListenerContext context = listener.GetContext();				
				HandleConnection (context);
			}
		}

		private IBroker GetBroker (string path)
		{
			IBroker best_match = null;
			int best_score = -1;

			foreach (IBroker broker in brokers) {
				if (Util.GlobMatch (broker.Path, path)) {
					int broker_score = broker.Path.Split ('/').Length;
					
					if (broker_score > best_score) {
						best_match = broker;
						best_score = broker_score;
					}
				}
			}
				
			return best_match;
		}
		
		private void HandleConnection (HttpListenerContext context)
		{				
			Log.Debug (this, "Got connection request ({0})", context.Request.RawUrl);
			//Log.Debug (this, context.User.Identity.Name);
			//Log.Debug (this, context.Request.Url.LocalPath);

			IBroker broker = GetBroker (context.Request.Url.LocalPath);
			
			if (broker == null)
				return;

			RequestWorker worker = new RequestWorker (broker, context);
			
			Thread worker_thread = new Thread (worker.DoWork);
			worker_thread.Start ();
		}

		private class RequestWorker {

			private IBroker broker = null;
			private HttpListenerContext context = null;
			
			public RequestWorker (IBroker broker, HttpListenerContext context)
			{
				this.broker = broker;
				this.context = context;
			}
			
			public void DoWork ()
			{
				Log.Debug (this, "Executing '{0}' broker", broker.Name);

				try {
					broker.HandleRequest (context);
				} catch (Exception e) {
					Log.Error (this, "Error executing broker '{0}':", broker.Name);
					Log.Error (this, e);
				}
			}
		}
	}
}

